import jax
import jax.numpy as jnp
import pytest

import blackjax.optimizers as optimizers


def test_dual_averaging():
    """We test the dual averaging algorithm by searching for the point that
    minimizes the gradient of a simple function.
    """

    # we need to wrap the gradient in a namedtuple as we optimize for a target
    # acceptance probability in the context of HMC.
    f = lambda x: (x - 1) ** 2

    # Our target gradient is 0. we increase the rate of convergence by
    # increasing the value of gamma (see documentation of the algorithm).
    init, update, final = optimizers.dual_averaging(gamma=0.3)

    da_state = init(3)
    for _ in range(100):
        x = jnp.exp(da_state.log_x)
        g = jax.grad(f)(x)
        da_state = update(da_state, g)

    assert final(da_state) == pytest.approx(1.0, 1e-1)



import jax
import numpy as np
import pytest

from blackjax.adaptation.mass_matrix import mass_matrix_adaptation


@pytest.mark.parametrize("n_dim", [1, 3])
@pytest.mark.parametrize("is_mass_matrix_diagonal", [True, False])
def test_welford_adaptation(n_dim, is_mass_matrix_diagonal):
    num_samples = 3_000

    # Generate samples from a multivariate normal distribution whose
    # covariance matrix we are going to try to recover.
    np.random.seed(0)
    mu = np.random.randn(n_dim)
    a = np.random.randn(n_dim, n_dim)
    cov = np.matmul(a.T, a)
    samples = np.random.multivariate_normal(mu, cov, num_samples)

    init, update, final = mass_matrix_adaptation(is_mass_matrix_diagonal)
    update_step = lambda state, sample: (update(state, sample), None)

    mm_state = init(n_dim)
    mm_state, _ = jax.lax.scan(update_step, mm_state, samples)
    mm_state = final(mm_state)

    estimated_cov = mm_state.inverse_mass_matrix

    if is_mass_matrix_diagonal:
        if n_dim == 1:
            np.testing.assert_allclose(estimated_cov, cov.squeeze(), rtol=1e-1)
        else:
            np.testing.assert_allclose(estimated_cov, np.diagonal(cov), rtol=1e-1)
    else:
        np.testing.assert_allclose(estimated_cov, cov, rtol=1e-1)



# Copyright Contributors to the Numpyro project.
# SPDX-License-Identifier: Apache-2.0
"""Test the iterative u-turn criterion."""
import jax.numpy as jnp
import pytest

from blackjax.inference.metrics import gaussian_euclidean
from blackjax.inference.termination import IterativeUTurnState, iterative_uturn_numpyro


@pytest.mark.parametrize(
    "checkpoint_idxs, expected_turning",
    [((3, 2), False), ((3, 3), True), ((0, 0), False), ((0, 1), True), ((1, 3), True)],
)
def test_is_iterative_turning(checkpoint_idxs, expected_turning):
    inverse_mass_matrix = jnp.ones(1)
    _, _, is_turning = gaussian_euclidean(inverse_mass_matrix)
    _, _, is_iterative_turning = iterative_uturn_numpyro(is_turning)

    momentum = 1.0
    momentum_sum = 3.0

    idx_min, idx_max = checkpoint_idxs
    momentum_ckpts = jnp.array([1.0, 2.0, 3.0, -2.0])
    momentum_sum_ckpts = jnp.array([2.0, 4.0, 4.0, -1.0])
    checkpoints = IterativeUTurnState(
        momentum_ckpts,
        momentum_sum_ckpts,
        idx_min,
        idx_max,
    )

    actual_turning = is_iterative_turning(checkpoints, momentum_sum, momentum)

    assert expected_turning == actual_turning



import jax
import jax.numpy as jnp
import numpy as np

import blackjax.hmc as hmc
from blackjax.adaptation.step_size import find_reasonable_step_size
from blackjax.inference.base import new_hmc_state


def test_reasonable_step_size():
    def potential_fn(x):
        return jnp.sum(0.5 * x)

    rng_key = jax.random.PRNGKey(0)

    init_position = jnp.array([3.0])
    reference_state = new_hmc_state(init_position, potential_fn)

    inv_mass_matrix = jnp.array([1.0])
    generator = lambda step_size: hmc.kernel(
        potential_fn, step_size, inv_mass_matrix, 10
    )

    # Test that the algorithm actually does something
    epsilon_1 = find_reasonable_step_size(
        rng_key,
        generator,
        reference_state,
        0.01,
        0.95,
    )
    assert epsilon_1 != 1.0
    assert epsilon_1 != np.inf

    # Different target acceptance rate
    epsilon_2 = find_reasonable_step_size(
        rng_key,
        generator,
        reference_state,
        1.0,
        0.05,
    )
    assert epsilon_2.item != epsilon_1



"""Base kernel for the HMC family."""
from typing import Callable, Dict, List, NamedTuple, Tuple, Union

import jax
import jax.numpy as jnp

from blackjax.inference.integrators import IntegratorState

__all__ = ["HMCState", "hmc"]

PyTree = Union[Dict, List, Tuple]


class HMCState(NamedTuple):
    """State of the HMC algorithm.

    The HMC algorithm takes one position of the chain and returns another
    position. In order to make computations more efficient, we also store
    the current potential energy as well as the current gradient of the
    potential energy.
    """

    position: PyTree
    potential_energy: float
    potential_energy_grad: PyTree


def new_hmc_state(position: PyTree, potential_fn: Callable) -> HMCState:
    """Create a chain state from a position.

    The HMC kernel works with states that contain the current chain position
    and its associated potential energy and derivative of the potential energy.
    This function computes initial states from an initial position.  While this
    function is intended to work for one chain, it is possible to use
    `jax.vmap` to compute the initial state of several chains.

    Note
    ----
    The potential energy is also known as the negative joint loglikelihood of
    the model, priors, and data without the normalizing constant, otherwise
    known as the negative log posterior density.

    Example
    -------
    Let us assume a model with two random variables. We wish to sample from
    4 chains with the following initial positions:

        >>> import numpy as np
        >>> init_positions = (np.random.rand(4, 1000), np.random.rand(4, 300))

    We have a `logpdf` function that returns the log-probability associated with
    the chain at a given position:

        >>> potential_fn((np.random.rand(1000), np.random.rand(3000)))
        -3.4

    We can compute the initial state for each of the 4 chain as follows:

        >>> import jax
        >>> jax.vmap(new_state, in_axes=(0, None))(init_positions, potential_fn)

    Parameters
    ----------
    position
        The current values of the random variables whose posterior we want to
        sample from. Can be anything from a list, a (named) tuple or a dict of
        arrays. The arrays can either be Numpy arrays or JAX DeviceArrays.
    potential_fn
        A function that returns the value of the potential energy when called
        with a position.

    Returns
    -------
    A HMC state that contains the position, the associated potential energy and gradient of the
    potential energy.
    """
    potential_energy, potential_energy_grad = jax.value_and_grad(potential_fn)(position)
    return HMCState(position, potential_energy, potential_energy_grad)


def hmc(
    momentum_generator: Callable,
    proposal_generator: Callable,
) -> Callable:
    """Create a Hamiltonian Monte Carlo transition kernel.

    Hamiltonian Monte Carlo (HMC) is known to yield effective Markov
    transitions and has been a major empirical success, leading to an extensive
    use in probabilistic programming languages and libraries [1,2,3]_.

    HMC works by augmenting the state space in which the chain evolves with an
    auxiliary momentum :math:`p`. At each step of the chain we draw a momentum
    value from the `momentum_generator` function. We then use Hamilton's
    equations [4]_ to push the state forward; we then compute the new state's
    energy using the `kinetic_energy` function and `logpdf` (potential energy).
    While the hamiltonian dynamics is conservative, numerical integration can
    introduce some discrepancy; we perform a Metropolis acceptance test to
    compensate for integration errors after having flipped the new state's
    momentum to make the transition reversible.

    I encourage anyone interested in the theoretical underpinning of the method
    to read Michael Betancourts' excellent introduction [3]_ and his more
    technical paper [5]_ on the geometric foundations of the method.

    This implementation is very general and should accomodate most variations
    on the method.

    Parameters
    ----------
    proposal_generator:
        The function used to propose a new state for the chain. For vanilla HMC this
        function integrates the trajectory over many steps, but gets more involved
        with other algorithms such as empirical and dynamical HMC.
    momentum_generator:
        A function that returns a new value for the momentum when called.
    kinetic_energy:
        A function that computes the current state's kinetic energy.
    divergence_threshold:
        The maximum difference in energy between the initial and final state
        after which we consider the transition to be divergent.

    Returns
    -------
    A kernel that moves the chain by one step when called.

    References
    ----------
    .. [1]: Duane, Simon, et al. "Hybrid monte carlo." Physics letters B
            195.2 (1987): 216-222.
    .. [2]: Neal, Radford M. "An improved acceptance procedure for the
            hybrid Monte Carlo algorithm." Journal of Computational Physics 111.1
            (1994): 194-203.
    .. [3]: Betancourt, Michael. "A conceptual introduction to
            Hamiltonian Monte Carlo." arXiv preprint arXiv:1701.02434 (2018).
    .. [4]: "Hamiltonian Mechanics", Wikipedia.
            https://en.wikipedia.org/wiki/Hamiltonian_mechanics#Deriving_Hamilton's_equations
    .. [5]: Betancourt, Michael, et al. "The geometric foundations
            of hamiltonian monte carlo." Bernoulli 23.4A (2017): 2257-2298.

    """

    def kernel(rng_key: jnp.ndarray, state: HMCState) -> Tuple[HMCState, NamedTuple]:
        """Moves the chain by one step using the Hamiltonian dynamics.

        Parameters
        ----------
        rng_key:
           The pseudo-random number generator key used to generate random numbers.
        state:
            The current state of the chain: position, log-probability and gradient
            of the log-probability.

        Returns
        -------
        The next state of the chain and additional information about the current step.
        """
        key_momentum, key_integrator = jax.random.split(rng_key, 2)

        position, potential_energy, potential_energy_grad = state
        momentum = momentum_generator(key_momentum, position)

        augmented_state = IntegratorState(
            position, momentum, potential_energy, potential_energy_grad
        )
        proposal, info = proposal_generator(key_integrator, augmented_state)
        proposal = HMCState(
            proposal.position, proposal.potential_energy, proposal.potential_energy_grad
        )

        return proposal, info

    return kernel



"""Test the trajectory integration"""
import functools

import jax
import jax.numpy as jnp
import numpy as np
import pytest

import blackjax.inference.integrators as integrators
import blackjax.inference.metrics as metrics
import blackjax.inference.proposal as proposal
import blackjax.inference.termination as termination
import blackjax.inference.trajectory as trajectory
from blackjax.inference.trajectory import DynamicExpansionState, Trajectory

divergence_threshold = 1000


@pytest.mark.parametrize("case", [(0.0001, False), (1000, True)])
def test_dynamic_progressive_integration_divergence(case):
    rng_key = jax.random.PRNGKey(0)

    def potential_fn(x):
        return jax.scipy.stats.norm.logpdf(x)

    step_size, should_diverge = case
    position = 1.0
    inverse_mass_matrix = jnp.array([1.0])

    momentum_generator, kinetic_energy_fn, uturn_check_fn = metrics.gaussian_euclidean(
        inverse_mass_matrix
    )

    integrator = integrators.velocity_verlet(potential_fn, kinetic_energy_fn)
    (
        new_criterion_state,
        update_criterion_state,
        is_criterion_met,
    ) = termination.iterative_uturn_numpyro(uturn_check_fn)

    trajectory_integrator = trajectory.dynamic_progressive_integration(
        integrator,
        kinetic_energy_fn,
        update_criterion_state,
        is_criterion_met,
        divergence_threshold,
    )

    # Initialize
    direction = 1
    initial_state = integrators.new_integrator_state(
        potential_fn, position, momentum_generator(rng_key, position)
    )
    initial_energy = initial_state.potential_energy + kinetic_energy_fn(
        initial_state.momentum
    )
    termination_state = new_criterion_state(initial_state, 10)
    max_num_steps = 100

    _, _, _, is_diverging, _ = trajectory_integrator(
        rng_key,
        initial_state,
        direction,
        termination_state,
        max_num_steps,
        step_size,
        initial_energy,
    )

    assert is_diverging.item() is should_diverge


def test_dynamic_progressive_equal_recursive():
    rng_key = jax.random.PRNGKey(23132)

    def potential_fn(x):
        return (1.0 - x[0]) ** 2 + 1.5 * (x[1] - x[0] ** 2) ** 2

    inverse_mass_matrix = jnp.asarray([[1.0, 0.5], [0.5, 1.25]])
    momentum_generator, kinetic_energy_fn, uturn_check_fn = metrics.gaussian_euclidean(
        inverse_mass_matrix
    )

    integrator = integrators.velocity_verlet(potential_fn, kinetic_energy_fn)
    (
        new_criterion_state,
        update_criterion_state,
        is_criterion_met,
    ) = termination.iterative_uturn_numpyro(uturn_check_fn)
    (
        integrator,
        kinetic_energy_fn,
        update_criterion_state,
        is_criterion_met,
        uturn_check_fn,
    ) = [
        jax.jit(x)
        for x in (
            integrator,
            kinetic_energy_fn,
            update_criterion_state,
            is_criterion_met,
            uturn_check_fn,
        )
    ]

    trajectory_integrator = trajectory.dynamic_progressive_integration(
        integrator,
        kinetic_energy_fn,
        update_criterion_state,
        is_criterion_met,
        divergence_threshold,
    )
    buildtree_integrator = trajectory.dynamic_recursive_integration(
        integrator,
        kinetic_energy_fn,
        uturn_check_fn,
        divergence_threshold,
    )

    for _ in range(50):
        (
            rng_key,
            rng_direction,
            rng_tree_depth,
            rng_step_size,
            rng_position,
            rng_momentum,
        ) = jax.random.split(rng_key, 6)
        direction = jax.random.choice(rng_direction, jnp.array([-1, 1]))
        tree_depth = jax.random.choice(rng_tree_depth, np.arange(2, 5))
        initial_state = integrators.new_integrator_state(
            potential_fn,
            jax.random.normal(rng_position, [2]),
            jax.random.normal(rng_momentum, [2]),
        )
        step_size = jnp.abs(jax.random.normal(rng_step_size, [])) * 0.1
        initial_energy = initial_state.potential_energy + kinetic_energy_fn(
            initial_state.momentum
        )

        termination_state = new_criterion_state(initial_state, tree_depth)
        (
            proposal0,
            trajectory0,
            _,
            is_diverging0,
            has_terminated0,
        ) = trajectory_integrator(
            rng_key,
            initial_state,
            direction,
            termination_state,
            2 ** tree_depth,
            step_size,
            initial_energy,
        )

        (
            _,
            proposal1,
            trajectory1,
            is_diverging1,
            has_terminated1,
        ) = buildtree_integrator(
            rng_key,
            initial_state,
            direction,
            tree_depth,
            step_size,
            initial_energy,
        )
        # Assert that the trajectory being built is the same
        jax.tree_multimap(
            functools.partial(np.testing.assert_allclose, rtol=1e-5),
            trajectory0,
            trajectory1,
        )
        assert is_diverging0 == is_diverging1
        assert has_terminated0 == has_terminated1
        # We dont expect the proposal to be the same (even with the same PRNGKey
        # as the order of selection is different). but the property associate
        # with the full trajectory should be the same.
        np.testing.assert_allclose(proposal0.weight, proposal1.weight, rtol=1e-5)
        np.testing.assert_allclose(
            proposal0.sum_log_p_accept, proposal1.sum_log_p_accept, rtol=1e-5
        )


@pytest.mark.parametrize(
    "case",
    [(0.0000000001, False, False, 10), (1, False, True, 2), (100000, True, True, 1)],
)
def test_dynamic_progressive_expansion(case):
    rng_key = jax.random.PRNGKey(0)

    def potential_fn(x):
        return 0.5 * x ** 2

    step_size, should_diverge, should_turn, expected_doublings = case
    position = 0.0
    inverse_mass_matrix = jnp.array([1.0])

    momentum_generator, kinetic_energy_fn, uturn_check_fn = metrics.gaussian_euclidean(
        inverse_mass_matrix
    )

    integrator = integrators.velocity_verlet(potential_fn, kinetic_energy_fn)
    (
        new_criterion_state,
        update_criterion_state,
        is_criterion_met,
    ) = termination.iterative_uturn_numpyro(uturn_check_fn)

    trajectory_integrator = trajectory.dynamic_progressive_integration(
        integrator,
        kinetic_energy_fn,
        update_criterion_state,
        is_criterion_met,
        divergence_threshold,
    )

    expand = trajectory.dynamic_multiplicative_expansion(
        trajectory_integrator, uturn_check_fn, step_size
    )

    state = integrators.new_integrator_state(
        potential_fn, position, momentum_generator(rng_key, position)
    )
    energy = state.potential_energy + kinetic_energy_fn(state.momentum)
    initial_proposal = proposal.Proposal(state, energy, 0.0, -np.inf)
    initial_termination_state = new_criterion_state(state, 10)
    initial_trajectory = Trajectory(
        state,
        state,
        state.momentum,
        0,
    )
    initial_expansion_state = DynamicExpansionState(
        0, initial_proposal, initial_trajectory, initial_termination_state
    )

    expansion_state, (is_diverging, is_turning) = expand(
        rng_key, initial_expansion_state, energy
    )

    assert is_diverging == should_diverge
    assert expansion_state.step == expected_doublings
    assert is_turning == should_turn



# Copyright Contributors to the Numpyro project.
# SPDX-License-Identifier: Apache-2.0
"""Make sure that the potential is only compiled once.
"""
import jax
import jax.numpy as jnp
import jax.scipy as jscipy

import blackjax.hmc as hmc
import blackjax.nuts as nuts

GLOBAL = {"count": 0}


def potential(x):
    GLOBAL["count"] += 1
    return jscipy.stats.norm.logpdf(x)


def test_hmc():
    """The reason why this works is because JAX only reads the potential once when compiled?"""
    rng_key = jax.random.PRNGKey(0)
    state = hmc.new_state(1.0, potential)

    GLOBAL["count"] = 0
    kernel = jax.jit(
        hmc.kernel(
            potential,
            step_size=1e-2,
            inverse_mass_matrix=jnp.array([1.0]),
            num_integration_steps=10,
        )
    )

    for _ in range(10):
        _, rng_key = jax.random.split(rng_key)
        state, _ = kernel(rng_key, state)

    assert GLOBAL["count"] == 1


def test_nuts():
    rng_key = jax.random.PRNGKey(0)
    state = hmc.new_state(1.0, potential)

    GLOBAL["count"] = 0
    kernel = jax.jit(
        nuts.kernel(potential, step_size=1e-2, inverse_mass_matrix=jnp.array([1.0]))
    )

    for _ in range(10):
        _, rng_key = jax.random.split(rng_key)
        state, _ = kernel(rng_key, state)

    # Potential function was traced twice as we call potential function
    # at Step 0 when building a new trajectory in tree doubling.
    assert GLOBAL["count"] == 2



"""Public API for the HMC Kernel"""
from typing import Callable, Dict, List, NamedTuple, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np

import blackjax.inference.base as base
import blackjax.inference.integrators as integrators
import blackjax.inference.metrics as metrics
import blackjax.inference.proposal as proposal
import blackjax.inference.trajectory as trajectory

Array = Union[np.ndarray, jnp.DeviceArray]
PyTree = Union[Dict, List, Tuple]

__all__ = ["new_state", "kernel"]


class HMCInfo(NamedTuple):
    """Additional information on the HMC transition.

    This additional information can be used for debugging or computing
    diagnostics.

    momentum:
        The momentum that was sampled and used to integrate the trajectory.
    acceptance_probability
        The acceptance probability of the transition, linked to the energy
        difference between the original and the proposed states.
    is_accepted
        Whether the proposed position was accepted or the original position
        was returned.
    is_divergent
        Whether the difference in energy between the original and the new state
        exceeded the divergence threshold.
    energy:
        Energy of the transition.
    proposal
        The state proposed by the proposal. Typically includes the position and
        momentum.
    step_size
        Size of the integration step.
    num_integration_steps
        Number of times we run the symplectic integrator to build the trajectory
    """

    momentum: PyTree
    acceptance_probability: float
    is_accepted: bool
    is_divergent: bool
    energy: float
    proposal: integrators.IntegratorState
    num_integration_steps: int


new_state = base.new_hmc_state


def kernel(
    potential_fn: Callable,
    step_size: float,
    inverse_mass_matrix: Array,
    num_integration_steps: int,
    divergence_threshold: int = 1000,
):
    """Build a HMC kernel.

    Parameters
    ----------
    potential_fn
        A function that returns the potential energy of a chain at a given position.
    parameters
        A NamedTuple that contains the parameters of the kernel to be built.

    Returns
    -------
    A kernel that takes a rng_key and a Pytree that contains the current state
    of the chain and that returns a new state of the chain along with
    information about the transition.

    """
    momentum_generator, kinetic_energy_fn, _ = metrics.gaussian_euclidean(
        inverse_mass_matrix
    )
    symplectic_integrator = integrators.velocity_verlet(potential_fn, kinetic_energy_fn)
    proposal_generator = hmc_proposal(
        symplectic_integrator,
        kinetic_energy_fn,
        step_size,
        num_integration_steps,
        divergence_threshold,
    )
    kernel = base.hmc(momentum_generator, proposal_generator)
    return kernel


def hmc_proposal(
    integrator: Callable,
    kinetic_energy: Callable,
    step_size: float,
    num_integration_steps: int = 1,
    divergence_threshold: float = 1000,
) -> Callable:
    """Vanilla HMC algorithm.

    The algorithm integrates the trajectory applying a symplectic integrator
    `num_integration_steps` times in one direction to get a proposal and uses a
    Metropolis-Hastings acceptance step to either reject or accept this
    proposal. This is what people usually refer to when they talk about "the
    HMC algorithm".

    Parameters
    ----------
    integrator
        Symplectic integrator used to build the trajectory step by step.
    kinetic_energy
        Function that computes the kinetic energy.
    step_size
        Size of the integration step.
    num_integration_steps
        Number of times we run the symplectic integrator to build the trajectory
    divergence_threshold
        Threshold above which we say that there is a divergence.

    Returns
    -------
    A kernel that generates a new chain state and information about the transition.

    """
    build_trajectory = trajectory.static_integration(
        integrator, step_size, num_integration_steps
    )
    init_proposal, generate_proposal = proposal.proposal_generator(
        kinetic_energy, divergence_threshold
    )
    sample_proposal = proposal.static_binomial_sampling

    def flip_momentum(
        state: integrators.IntegratorState,
    ) -> integrators.IntegratorState:
        """To guarantee time-reversibility (hence detailed balance) we
        need to flip the last state's momentum. If we run the hamiltonian
        dynamics starting from the last state with flipped momentum we
        should indeed retrieve the initial state (with flipped momentum).

        """
        flipped_momentum = jax.tree_util.tree_multimap(
            lambda m: -1.0 * m, state.momentum
        )
        return integrators.IntegratorState(
            state.position,
            flipped_momentum,
            state.potential_energy,
            state.potential_energy_grad,
        )

    def generate(
        rng_key, state: integrators.IntegratorState
    ) -> Tuple[integrators.IntegratorState, HMCInfo]:
        """Generate a new chain state."""
        end_state = build_trajectory(state)
        end_state = flip_momentum(end_state)
        proposal = init_proposal(state)
        new_proposal, is_diverging = generate_proposal(proposal.energy, end_state)
        sampled_proposal, *info = sample_proposal(rng_key, proposal, new_proposal)
        do_accept, p_accept = info

        info = HMCInfo(
            state.momentum,
            p_accept,
            do_accept,
            is_diverging,
            new_proposal.energy,
            new_proposal,
            num_integration_steps,
        )

        return sampled_proposal.state, info

    return generate



import jax
import jax.numpy as jnp
import numpy
import pytest

import blackjax.inference.integrators as integrators


def HarmonicOscillator(inv_mass_matrix, k=1.0, m=1.0):
    """Potential and Kinetic energy of an harmonic oscillator."""

    def potential_energy(q):
        return jnp.sum(0.5 * k * jnp.square(q["x"]))

    def kinetic_energy(p):
        v = jnp.multiply(inv_mass_matrix, p["x"])
        return jnp.sum(0.5 * jnp.dot(v, p["x"]))

    return potential_energy, kinetic_energy


def FreeFall(inv_mass_matrix, g=1.0):
    """Potential and kinetic energy of a free-falling object."""

    def potential_energy(q):
        return jnp.sum(g * q["x"])

    def kinetic_energy(p):
        v = jnp.multiply(inv_mass_matrix, p["x"])
        return jnp.sum(0.5 * jnp.dot(v, p["x"]))

    return potential_energy, kinetic_energy


def PlanetaryMotion(inv_mass_matrix):
    """Potential and kinetic energy for planar planetary motion."""

    def potential_energy(q):
        return -1.0 / jnp.power(q["x"] ** 2 + q["y"] ** 2, 0.5)

    def kinetic_energy(p):
        z = jnp.stack([p["x"], p["y"]], axis=-1)
        return 0.5 * jnp.dot(inv_mass_matrix, z ** 2)

    return potential_energy, kinetic_energy


algorithms = [{"integrator": integrators.velocity_verlet, "precision": 1e-4}]


examples = [
    {
        "model": FreeFall,
        "num_steps": 100,
        "step_size": 0.01,
        "q_init": {"x": 0.0},
        "p_init": {"x": 1.0},
        "q_final": {"x": 0.5},
        "p_final": {"x": 1.0},
        "inv_mass_matrix": jnp.array([1.0]),
    },
    {
        "model": HarmonicOscillator,
        "num_steps": 100,
        "step_size": 0.01,
        "q_init": {"x": 0.0},
        "p_init": {"x": 1.0},
        "q_final": {"x": jnp.sin(1.0)},
        "p_final": {"x": jnp.cos(1.0)},
        "inv_mass_matrix": jnp.array([1.0]),
    },
    {
        "model": PlanetaryMotion,
        "num_steps": 628,
        "step_size": 0.01,
        "q_init": {"x": 1.0, "y": 0.0},
        "p_init": {"x": 0.0, "y": 1.0},
        "q_final": {"x": 1.0, "y": 0.0},
        "p_final": {"x": 0.0, "y": 1.0},
        "inv_mass_matrix": jnp.array([1.0, 1.0]),
    },
]


@pytest.mark.parametrize("do_jit_compile", [True, False])
@pytest.mark.parametrize("integrator", algorithms)
@pytest.mark.parametrize("example", examples)
def test_integrator(example, integrator, do_jit_compile):
    """Test the numerical accuracy of trajectory integrators.

    We compare the evolution of the trajectory to analytical integration, and
    the conservation of energy. JAX's default float precision is 32bit; it is
    possible to change it to 64bit but only at startup. It is thus impossible
    to test both in the same run; we run the tests with the lower precision.

    """
    model = example["model"]
    potential, kinetic_energy = model(example["inv_mass_matrix"])
    integrator_step = integrator["integrator"]

    step = integrator_step(potential, kinetic_energy)
    step = jax.jit(step) if do_jit_compile else step

    step_size = example["step_size"]

    q = example["q_init"]
    p = example["p_init"]
    initial_state = integrators.IntegratorState(
        q, p, potential(q), jax.grad(potential)(q)
    )
    final_state = jax.lax.fori_loop(
        0, example["num_steps"], lambda _, state: step(state, step_size), initial_state
    )

    # We make sure that the particle moved from its initial position.
    for dimension in final_state.position:
        init = final_state.position[dimension]
        final = example["q_final"][dimension]
        print(example["q_final"], final_state.position)
        numpy.testing.assert_allclose(init, final, atol=1e-2)

    # We now check the conservation of energy, the property that matters the most in HMC.
    energy = potential(q) + kinetic_energy(p)
    new_energy = potential(final_state.position) + kinetic_energy(final_state.momentum)
    assert energy == pytest.approx(new_energy.item(), integrator["precision"])



"""Test the accuracy of the HMC kernel"""
import functools as ft

import jax
import jax.numpy as jnp
import jax.scipy.stats as stats
import numpy as np
import pytest

import blackjax.hmc as hmc
import blackjax.nuts as nuts
import blackjax.stan_warmup as stan_warmup


def inference_loop(rng_key, kernel, initial_state, num_samples):
    def one_step(state, rng_key):
        state, _ = kernel(rng_key, state)
        return state, state

    keys = jax.random.split(rng_key, num_samples)
    _, states = jax.lax.scan(one_step, initial_state, keys)

    return states


# -------------------------------------------------------------------
#                        LINEAR REGRESSION
# -------------------------------------------------------------------


def regression_potential_fn(scale, coefs, preds, x):
    """Linear regression"""
    logpdf = 0
    logpdf += stats.expon.logpdf(scale, 1, 1)
    logpdf += stats.norm.logpdf(coefs, 3 * jnp.ones(x.shape[-1]), 2)
    y = jnp.dot(x, coefs)
    logpdf += stats.norm.logpdf(preds, y, scale)
    return -jnp.sum(logpdf)


regresion_test_cases = [
    {
        "algorithm": hmc,
        "initial_position": {"scale": 1.0, "coefs": 2.0},
        "parameters": {"num_integration_steps": 90},
        "num_warmup_steps": 3_000,
        "num_sampling_steps": 2_000,
    },
    {
        "algorithm": nuts,
        "initial_position": {"scale": 1.0, "coefs": 2.0},
        "parameters": {},
        "num_warmup_steps": 1_000,
        "num_sampling_steps": 500,
    },
]


@pytest.mark.parametrize("case", regresion_test_cases)
@pytest.mark.parametrize("is_mass_matrix_diagonal", [True, False])
def test_linear_regression(case, is_mass_matrix_diagonal):
    """Test the HMC kernel and the Stan warmup."""
    x_data = np.random.normal(0, 1, size=(1000, 1))
    y_data = 3 * x_data + np.random.normal(size=x_data.shape)
    observations = {"x": x_data, "preds": y_data}

    conditioned_potential = ft.partial(regression_potential_fn, **observations)
    potential = lambda x: conditioned_potential(**x)

    rng_key = jax.random.PRNGKey(19)
    initial_position = case["initial_position"]
    initial_state = case["algorithm"].new_state(initial_position, potential)

    kernel_factory = lambda step_size, inverse_mass_matrix: case["algorithm"].kernel(
        potential, step_size, inverse_mass_matrix, **case["parameters"]
    )

    state, (step_size, inverse_mass_matrix), _ = stan_warmup.run(
        rng_key,
        kernel_factory,
        initial_state,
        case["num_warmup_steps"],
        is_mass_matrix_diagonal=is_mass_matrix_diagonal,
    )

    if is_mass_matrix_diagonal:
        assert inverse_mass_matrix.ndim == 1
    else:
        assert inverse_mass_matrix.ndim == 2

    kernel = kernel_factory(step_size, inverse_mass_matrix)
    states = inference_loop(rng_key, kernel, initial_state, case["num_sampling_steps"])

    coefs_samples = states.position["coefs"]
    scale_samples = states.position["scale"]

    assert np.mean(scale_samples) == pytest.approx(1, 1e-1)
    assert np.mean(coefs_samples) == pytest.approx(3, 1e-1)


# -------------------------------------------------------------------
#                UNIVARIATE NORMAL DISTRIBUTION
# -------------------------------------------------------------------


def normal_potential_fn(x):
    return -stats.norm.logpdf(x, loc=1.0, scale=2.0).squeeze()


normal_test_cases = [
    {
        "algorithm": hmc,
        "initial_position": {"x": 1.0},
        "parameters": {
            "step_size": 1e-2,
            "inverse_mass_matrix": jnp.array([0.1]),
            "num_integration_steps": 100,
        },
        "num_sampling_steps": 50_000,
    },
    {
        "algorithm": nuts,
        "initial_position": {"x": 1.0},
        "parameters": {"step_size": 1e-2, "inverse_mass_matrix": jnp.array([0.1])},
        "num_sampling_steps": 50_000,
    },
]


@pytest.mark.parametrize("case", normal_test_cases)
@pytest.mark.parametrize("is_mass_matrix_diagonal", [True, False])
def test_univariate_normal(case, is_mass_matrix_diagonal):
    rng_key = jax.random.PRNGKey(19)
    potential = lambda x: normal_potential_fn(**x)
    initial_position = case["initial_position"]
    initial_state = case["algorithm"].new_state(initial_position, potential)

    kernel = case["algorithm"].kernel(potential, **case["parameters"])
    states = inference_loop(rng_key, kernel, initial_state, case["num_sampling_steps"])

    samples = states.position["x"]

    assert np.var(samples).item() == pytest.approx(4.0, 1e-1)
    assert np.mean(samples).item() == pytest.approx(1.0, 1e-1)



import jax.numpy as jnp
import pytest
from jax import random

from blackjax.inference import metrics

KEY = random.PRNGKey(0)
DTYPE = "float32"


@pytest.mark.parametrize("shape", [(), (1, 2, 3)], ids=lambda x: f"shape {x}")
def test_gaussian_euclidean_ndim_invalid(shape):
    """Test Gaussian Euclidean Function returns correct function invalid ndim"""
    x = jnp.ones(shape=shape)

    with pytest.raises(ValueError) as e:
        metrics.gaussian_euclidean(x)
    assert "The mass matrix has the wrong number of dimensions" in str(e)


def test_gaussian_euclidean_dim_1():
    """Test Gaussian Euclidean Function with ndim 1"""
    inverse_mass_matrix = jnp.asarray([1 / 4], dtype=DTYPE)
    momentum, kinetic_energy, _ = metrics.gaussian_euclidean(inverse_mass_matrix)

    arbitrary_position = jnp.asarray([12345], dtype=DTYPE)
    momentum_val = momentum(KEY, arbitrary_position)

    # 2 is square root inverse of 1/4
    # -0.20584235 is random value returned with random key
    expected_momentum_val = 2 * -0.20584235

    kinetic_energy_val = kinetic_energy(momentum_val)
    velocity = inverse_mass_matrix * momentum_val
    expected_kinetic_energy_val = 0.5 * velocity * momentum_val

    assert momentum_val == expected_momentum_val
    assert kinetic_energy_val == expected_kinetic_energy_val


def test_gaussian_euclidean_dim_2():
    """Test Gaussian Euclidean Function with ndim 2"""
    inverse_mass_matrix = jnp.asarray([[1 / 9, 0], [0, 1 / 4]], dtype=DTYPE)
    momentum, kinetic_energy, _ = metrics.gaussian_euclidean(inverse_mass_matrix)

    arbitrary_position = jnp.asarray([12345, 23456], dtype=DTYPE)
    momentum_val = momentum(KEY, arbitrary_position)

    # 2 is square root inverse of 1/4
    # -0.20584235 is random value returned with random key
    expected_momentum_val = jnp.asarray([3, 2]) * jnp.asarray([-0.784766, 0.8564448])

    kinetic_energy_val = kinetic_energy(momentum_val)
    velocity = jnp.dot(inverse_mass_matrix, momentum_val)
    expected_kinetic_energy_val = 0.5 * jnp.matmul(velocity, momentum_val)

    assert pytest.approx(expected_momentum_val, momentum_val)
    assert kinetic_energy_val == expected_kinetic_energy_val
